/*
  Maxwell's Lemur
  Copyright (c) 1997 Daniel Risacher <magnus@alum.mit.edu>
  
  Siag, Scheme In A Grid
  Copyright (C) 1996, 1997  Ulric Eriksson <ulric@edu.stockholm.se>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * interpreters.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <gtk/gtk.h>
#include "core.h"

#include "matrix.h"

gint errorflag;
ValueType siag_type;

#define MAX_INTERPRETERS 10

gint nent_interpreters;

static struct {
  gchar *name;
  cval (*evalfn)();
  void (*execfn)();
  gchar *(*ref_fn)();
} interpreters[MAX_INTERPRETERS];

static int ref_finder(gchar *src, gchar *brk, 
		      gchar **bname, gchar **bpname, 
		      gchar **rc_begin, gchar **rc_end,
		      gint *row, gint *col) ;


gchar *interpreter2name(gint interpreter)
{
  if (interpreter < 0 || interpreter >= nent_interpreters)
    return "NONE";
  return interpreters[interpreter].name;
}

gint name2interpreter(gchar *name)
{
  gint i;
  for (i = 0; i < nent_interpreters; i++)
    if (!strcasecmp(name, interpreters[i].name)) return i;
  return -1;
}

gint register_interpreter(gchar *name, cval (*evalfn)(), void (*execfn)(),
			  gchar *(*ref_fn)())
{
  if (nent_interpreters >= MAX_INTERPRETERS) return -1;	
  interpreters[nent_interpreters].evalfn = evalfn;
  interpreters[nent_interpreters].execfn = execfn;
  interpreters[nent_interpreters].ref_fn = ref_fn;
  interpreters[nent_interpreters].name = name;
  return nent_interpreters++;
}

void exec_expr(gint interpreter, gchar *expr)
{
  void (*execfn)();
  
  if (interpreter < 0 || interpreter >= nent_interpreters) return;
  execfn = interpreters[interpreter].execfn;
  if (execfn == NULL) {
#ifdef DEBUG
    fprintf(stderr, "No executing %s commands\n",
	    interpreters[interpreter].name);
#endif
    return;
  }
#ifdef DEBUG
  printf("executing %s command \"%s\"\n",
	 interpreters[interpreter].name, expr);
#endif
  execfn(expr);
}

/* This same function also does strings */
cval parse_expr(buffer *b, buffer_page *bp, gint interpreter,
		gchar *expr, gint row, gint col)
{
  cval (*evalfn)();

  /* the next two lines seem to contradict each other, but it is
     necessary to set the type in case siod bails out before
     we make it to siag_print(). */
  errorflag = 0;
  siag_type = VALUE_TYPE_ERROR;
  
  if (interpreter >= nent_interpreters) return 0;
  
  evalfn = interpreters[interpreter].evalfn;
  if (evalfn == NULL) return 0;
#ifdef DEBUG
  printf("evaluating %s expression \"%s\"\n",
	 interpreters[interpreter].name, expr);
#endif
  
  return evalfn(b, bp, expr, row, col);
}

/*
   int calc_matrix(spread **matrix)
   Recalculate all the cells in matrix.
   Returns 0 if successful, otherwise 1.
*/
/* Remarkably enough, this function requires no conversion whatsoever
   to go from long to double. */
/* FIXME */
gint calc_matrix(buffer *b)
{
  buffer_page *bp;
  gint row, col;
  gint lr, lc;
  CellType ctype;
  
  errorflag = FALSE;
  for (bp = b->pages; bp; bp = bp->next) {
    lr = row_last_used(bp);
    for (row = 1; row <= lr; row++) {
      lc = col_last_used(bp, row);
      for (col = 1; col <= lc; col++) {
	switch (ctype = ret_ctype(bp, row, col)) {
	  cval value;
	case CELL_TYPE_STRING:
	case CELL_TYPE_EXPRESSION:
	  value = parse_expr(b, bp,
			     ret_interpreter(bp, row, col),
			     ret_taddr(bp, row, col),
			     row, col);
	  ins_data(bp,
		   ret_interpreter(bp, row, col),
		   ret_formula(bp, row, col), ctype,
		   value, siag_type, row, col);
	  break;
	default:
	  ;
	}
      }
    }
  }
  return 0;
}				/* calc_matrix */

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

/* make a rough upper bound on the number of references */
gint ref_counter(gchar *expr)
{
  gint cd, rd, c;
  
  cd = 0;
  rd = 0;
  
  while ((c = *expr++)) {
    if (c == 'r' || c == 'R') {
      if (isdigit(*expr)) rd++;
    } else if (c == 'c' || c == 'C') {
      if (isdigit(*expr)) cd++;
    }
  }
  
  return MIN(cd, rd);
}


/* ref_expander()
   
   Recognizes strings in src of the form "foo:bar:r1c2" and replaces them 
   with a filled in template, such as "(get-cell %ld %ld \"%s\" \"%s\")"
   It also recognizes "foo:r1c1" and "r1c1", passing an empty string
   for the second & third arguments to the template, as appropriate */

typedef enum {
  ASSUME_RnCn, ASSUME_NAME, KNOWN_STRING
} ref_state;


void ref_expander(gchar *src, gchar *dest, gchar *brk, gchar *tmpl) 
{
  gchar *bname, *bpname, *rc0, *rc1;
  gchar *first;
  gint row, col;
  gint found;
  gchar bname_copy[50];  /*FIXME*/
  gchar bpname_copy[50]; /*FIXME*/

  while ((found = ref_finder(src, brk,
			    &bname, &bpname, &rc0, &rc1,
			    &row, &col))) {

    first = bname?bname:(bpname?bpname:rc0);
    while (src < first) 
      *dest++ = *src++;
    src = rc1;
    if (bpname) { 
      strncpy(bpname_copy,bpname,rc0-bpname); 
      bpname_copy[rc0-bpname-1] = 0;
    } else {
      bpname_copy[0] = 0;
    }
    if (bname)  { 
      strncpy(bname_copy,bname,bpname-bname); 
      bname_copy[bpname-bname-1] = 0;
    } else {
      bname_copy[0] = 0;
    }
    dest += sprintf(dest, tmpl, row, col, bpname_copy, bname_copy);
  }
  while ((*dest++ = *src++));
}

/* FIXME: this should know about other pages, & other buffers */
/* move block between (r1,c1) and (r2,c2), direction (rd,cd) */
void ref_updater(gchar *src, gchar *dest, gchar *brk,
		 gint r1, gint c1, gint r2, gint c2, gint rd, gint cd)
{
  gchar *bname, *bpname, *rc0, *rc1;
  gchar *first;
  gint row, col;
  gint found;

  while ((found = ref_finder(src, brk,
			    &bname, &bpname, &rc0, &rc1,
			    &row, &col))) {
    first = bname?bname:(bpname?bpname:rc0);
    while (src < rc0) 
      *dest++ = *src++;
    src = rc1;
    if (row >= r1 && row <= r2 &&
	col >= c1 && col <= c2) {
      row += rd;
      col += cd;
    }
    dest += sprintf(dest,"R%dC%d",row,col);
  }
  while ((*dest++ = *src++));
}

/* ref_finder()
   finds references of form "buffer:bpage:RnCn", "bpage:RnCn", or "RnCn"
*/
/* works like a FSM: always assume that the upcoming input is a
   "r1c1"-ref unless we can prove otherwise.  If we can, assume it's a
   page name, unless we can prove otherwise.  If we've already got a
   page name, assume the first thing was really a buffer name.  If it
   ain't none a' does tings, pass it to the output unchanged.

   start_assumption is the start of the text we're still looking at

   interim_assumption is the start of the text we're looking at but
   haven't categorized yet

   end_assumption is (just past) the end of the text we've looked at */

static gint ref_finder(gchar *src, gchar *brk, 
		       gchar **bname, gchar **bpname, 
		       gchar **rc_begin, gchar **rc_end,
		       gint *row, gint *col) 
{
  ref_state state = ASSUME_RnCn;
  gchar *start_assumption, *interim_assumption, *end_assumption;
  gint c;

  *bname = *bpname = NULL;

  start_assumption = interim_assumption = end_assumption = src;
  while ((c = *end_assumption++)) {
    switch(state) {
    case KNOWN_STRING:
      if (c == '"')  { 	
	start_assumption = interim_assumption = end_assumption;
	state = ASSUME_RnCn; 
      } 
      if (c == '\\') {             /* consume next character also */
	end_assumption++; 
      }
      break; 

    case ASSUME_RnCn:
      if (c == '"') { state = KNOWN_STRING;  break; }

      if ((c == 'r' || c == 'R') && isdigit(*end_assumption)) {
	/* found component 1&2: "Rn" */
/* 	if (*bpname == NULL) *bpname = interim_assumption; */
/* 	if (*bname == NULL)  *bname = *bpname;  */
	*rc_begin = interim_assumption;
	*row = strtol(end_assumption, &end_assumption, 10);
	c = *end_assumption++;  /* FIXME: end of string? */
	if ((c == 'c' || c == 'C') && isdigit(*end_assumption)) {
	  /* found component 3&4: "Cn" */
	  *col = strtol(end_assumption, &end_assumption, 10);
	  if (*end_assumption && !strchr(brk, *end_assumption)) {
	    /* didn't end with a break character: "RnCnSOMETHING" */
	    state = ASSUME_NAME;
	  } else {
	    /* found all 5 conditions: "RnCn " ! */
	    *rc_end = end_assumption;
	    return 1;
	  }
	}
      } else { 	/* didn't find "Rn" */
	state = ASSUME_NAME;
      }
    case ASSUME_NAME:
      if (c == ':') {                    /* end of a bname or bpname */
	if (*bpname == NULL) {  /* page name if we don't have one yet */
	  *bpname = interim_assumption;
	  interim_assumption = end_assumption;
      } else {                        /* page name if we do have one */
	  *bname = *bpname; 
	  *bpname = interim_assumption;
	  interim_assumption = end_assumption;
      }
	state = ASSUME_RnCn;         
      } else if (strchr(brk, c)) {
	/* break char --> wasn't a name, so punt */
	start_assumption = interim_assumption = end_assumption;
	state = ASSUME_RnCn;
      } else {
	/* wasn't a ':' or a brk, so keep going */
	break;
      }
    }
  }
  /* must not have found any reference */
  bname = bpname = rc_end = rc_begin = NULL;
  return 0;
}

/* expand Visicalc references */

gchar *update_references(gint interpreter, gchar *expr,
			 gint r1, gint c1, gint r2, gint c2, gint rd, gint cd)
{
  char *(*ref_fn)();
  
  if (interpreter < 0 || interpreter >= nent_interpreters) return expr;
  ref_fn = interpreters[interpreter].ref_fn;
  if (ref_fn == NULL) {
    return expr;
  }
  return ref_fn(expr, r1, c1, r2, c2, rd, cd);
}

/*FIXME*/
void update_all_references(buffer *b, buffer_page *bp, 
			   gint r1, gint c1, gint r2, gint c2,
			   gint rd, gint cd)
{
  return;
}

void init_interpreters()
{
  nent_interpreters = 0;
}

